@@ -767,7 +767,7 @@ bm_menu_set_selected_items(struct bm_menu *menu, struct bm_item **items, uint32_
     assert(menu);
 
     struct bm_item **new_items;
-    if (!(new_items = calloc(sizeof(struct bm_item*), nmemb)))
+    if (!(new_items = calloc(nmemb, sizeof(struct bm_item*))))
         return 0;
 
     memcpy(new_items, items, sizeof(struct bm_item*) * nmemb);
@@ -1084,19 +1084,35 @@ bm_menu_run_with_key(struct bm_menu *menu, enum bm_key key, uint32_t unicode)
             break;
 
         case BM_KEY_PAGE_UP:
-            bm_menu_set_highlighted_index(menu, (menu->index < displayed ? 0 : menu->index - (displayed - 1)));
+            if (menu->lines_mode == BM_LINES_UP) {
+                bm_menu_set_highlighted_index(menu, (menu->index + displayed >= count ? count - 1 : menu->index + (displayed - 1)));
+            } else {
+                bm_menu_set_highlighted_index(menu, (menu->index < displayed ? 0 : menu->index - (displayed - 1)));
+            }
             break;
 
         case BM_KEY_PAGE_DOWN:
-            bm_menu_set_highlighted_index(menu, (menu->index + displayed >= count ? count - 1 : menu->index + (displayed - 1)));
+            if (menu->lines_mode == BM_LINES_UP) {
+                bm_menu_set_highlighted_index(menu, (menu->index < displayed ? 0 : menu->index - (displayed - 1)));
+            } else {
+                bm_menu_set_highlighted_index(menu, (menu->index + displayed >= count ? count - 1 : menu->index + (displayed - 1)));
+            }
             break;
 
         case BM_KEY_SHIFT_PAGE_UP:
-            bm_menu_set_highlighted_index(menu, 0);
+            if (menu->lines_mode == BM_LINES_UP) {
+                bm_menu_set_highlighted_index(menu, count - 1);
+            } else {
+                bm_menu_set_highlighted_index(menu, 0);
+            }
             break;
 
         case BM_KEY_SHIFT_PAGE_DOWN:
-            bm_menu_set_highlighted_index(menu, count - 1);
+            if (menu->lines_mode == BM_LINES_UP) {
+                bm_menu_set_highlighted_index(menu, 0);
+            } else {
+                bm_menu_set_highlighted_index(menu, count - 1);
+            }
             break;
 
         case BM_KEY_BACKSPACE:
@@ -1150,9 +1166,18 @@ bm_menu_run_with_key(struct bm_menu *menu, enum bm_key key, uint32_t unicode)
             }
             break;
 
-        case BM_KEY_PASTE:
+        case BM_KEY_PASTE_PRIMARY:
+        case BM_KEY_PASTE_CLIPBOARD:
             {
-                char *paster[] = {"wl-paste -t text/plain", "xclip -t text/plain -out"};
+                char *paster[2] = {NULL, NULL};
+                if (key == BM_KEY_PASTE_PRIMARY) {
+                    paster[0] = "wl-paste -t text -p";
+                    paster[1] = "xclip -t UTF8_STRING -out";
+                } else {
+                    paster[0] = "wl-paste -t text";
+                    paster[1] = "xclip -t UTF8_STRING -out -selection clipboard";
+                }
+
                 for (size_t paster_i = 0; paster_i < sizeof paster / sizeof paster[0]; paster_i++) {
                     FILE *clipboard = popen(paster[paster_i], "r");
                     if (clipboard == NULL) {
@@ -1390,7 +1415,6 @@ bm_menu_run_with_touch(struct bm_menu *menu, struct bm_touch touch)
     }
 
     if (count_down == 2) {
-        int16_t scroll_count = 0;
         int16_t scroll_directions[2];
         int16_t distance_trigger = displayed * bm_menu_get_line_height(menu) / 4;
         for (size_t i = 0; i < 2; ++i) {
@@ -1401,7 +1425,6 @@ bm_menu_run_with_touch(struct bm_menu *menu, struct bm_touch touch)
             int32_t movement_y = point.pos_y - point.start_y;
             if (abs(movement_y) > distance_trigger) {
                 scroll_directions[i] = movement_y / abs(movement_y);
-                scroll_count++;
             }
         }
 
