@@ -13,6 +13,7 @@ struct cairo {
     cairo_surface_t *surface;
     PangoContext *pango;
     double scale;
+    bool antialiasing;
 };
 
 struct cairo_color {
@@ -67,7 +68,8 @@ bm_cairo_create_for_surface(struct cairo *cairo, cairo_surface_t *surface)
     if (!(cairo->pango = pango_cairo_create_context(cairo->cr)))
         goto fail;
 
-    cairo_set_antialias(cairo->cr, CAIRO_ANTIALIAS_NONE);
+    if (!cairo->antialiasing)
+        cairo_set_antialias(cairo->cr, CAIRO_ANTIALIAS_NONE);
 
     cairo->surface = surface;
     assert(cairo->scale > 0);
@@ -85,6 +87,8 @@ bm_cairo_destroy(struct cairo *cairo)
 {
     if (cairo->cr)
         cairo_destroy(cairo->cr);
+    if (cairo->pango)
+        g_object_unref(cairo->pango);
     if (cairo->surface)
         cairo_surface_destroy(cairo->surface);
 }
@@ -130,6 +134,7 @@ bm_pango_get_text_extents(struct cairo *cairo, struct cairo_paint *paint, struct
 static inline bool
 bm_cairo_draw_line_str(struct cairo *cairo, struct cairo_paint *paint, struct cairo_result *result, const char *buffer)
 {
+    cairo_save(cairo->cr);
     PangoLayout *layout = bm_pango_get_layout(cairo, paint, buffer);
     pango_cairo_update_layout(cairo->cr, layout);
 
@@ -188,6 +193,7 @@ bm_cairo_draw_line_str(struct cairo *cairo, struct cairo_paint *paint, struct ca
     result->height = line_height;
 
     cairo_identity_matrix(cairo->cr);
+    cairo_restore(cairo->cr);
     return true;
 }
 
@@ -260,17 +266,19 @@ bm_cairo_entry_message(char *entry_text, bool highlighted, uint32_t event_feedba
 }
 
 static inline void
-bm_cairo_draw_rounded_path(cairo_t *cr, double x, double y, double width, double height, double radius)
+bm_cairo_rounded_path(cairo_t *cr, double x, double y, double width, double height, double radius)
 {
+    double adjusted_radius = MIN(MIN(radius, height * 0.5), width * 0.5); /* Prevent border from intersecting itself. */
     double degrees = M_PI / 180;
     cairo_new_sub_path(cr);
-    cairo_arc(cr, x + width - radius, y + radius, radius, -90 * degrees, 0 * degrees);
-    cairo_arc(cr, x + width - radius, y + height - radius, radius, 0 * degrees, 90 * degrees);
-    cairo_arc(cr, x + radius, y + height - radius, radius, 90 * degrees, 180 * degrees);
-    cairo_arc(cr, x + radius, y + radius, radius, 180 * degrees, 270 * degrees);
+    cairo_arc(cr, x + width - adjusted_radius, y + adjusted_radius, adjusted_radius, -90 * degrees, 0 * degrees);
+    cairo_arc(cr, x + width - adjusted_radius, y + height - adjusted_radius, adjusted_radius, 0 * degrees, 90 * degrees);
+    cairo_arc(cr, x + adjusted_radius, y + height - adjusted_radius, adjusted_radius, 90 * degrees, 180 * degrees);
+    cairo_arc(cr, x + adjusted_radius, y + adjusted_radius, adjusted_radius, 180 * degrees, 270 * degrees);
     cairo_close_path(cr);
 }
 
+
 static inline void
 bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct bm_menu *menu, struct cairo_paint_result *out_result)
 {
@@ -302,11 +310,9 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct
     uint32_t filtered_item_count;
     bm_menu_get_filtered_items(menu, &filtered_item_count);
 
-    cairo_set_source_rgba(cairo->cr, 0, 0, 0, 0);
-    cairo_rectangle(cairo->cr, 0, 0, width, height);
-
     cairo_save(cairo->cr);
     cairo_set_operator(cairo->cr, CAIRO_OPERATOR_CLEAR);
+    cairo_reset_clip(cairo->cr);
     cairo_paint(cairo->cr);
     cairo_restore(cairo->cr);
 
@@ -323,6 +329,16 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct
     memset(&result, 0, sizeof(result));
     uint32_t title_x = 0;
 
+    uint32_t total_height = menu->lines * height;
+    if (!menu->fixed_height) total_height = MIN(count * height, total_height);
+    total_height += height +border_size * 2;
+    total_height = MIN(total_height, (max_height / height) * height);
+
+    if (border_radius) {
+        bm_cairo_rounded_path(cairo->cr, 1, 1, ((width + border_size) / cairo->scale) - 2, total_height - 2, border_radius);
+        cairo_clip(cairo->cr);
+    }
+
     if (menu->title) {
         /* Essentially hide the title text if we are drawing lines "up". */
         enum bm_color title_fg = (menu->lines_mode == BM_LINES_UP ? BM_COLOR_ITEM_BG : BM_COLOR_TITLE_FG);
@@ -347,7 +363,7 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct
     paint.cursor = menu->cursor;
     paint.cursor_height = menu->cursor_height;
     paint.cursor_width = menu->cursor_width;
-    paint.pos = (struct pos){ (menu->title ? 2 : 0) + result.x_advance + border_size, vpadding + border_size };
+    paint.pos = (struct pos){ (menu->title ? 6 : 0) + result.x_advance + border_size, vpadding + border_size };
     paint.box = (struct box){ (menu->title ? 2 : 4), 0, vpadding, -vpadding, width - paint.pos.x, height };
 
     if (menu->lines == 0 || menu->lines_mode == BM_LINES_DOWN) {
@@ -356,9 +372,7 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct
             bm_cairo_draw_line_str(cairo, &paint, &result, "");
         } else if (menu->password == BM_PASSWORD_INDICATOR) {
             char asterisk_print[1024] = "";
-        
             for (int i = 0; i < (int)(strlen(filter_text)); ++i) asterisk_print[i] = '*';
-        
             bm_cairo_draw_line(cairo, &paint, &result, "%s", asterisk_print);
         } else if (menu->password == BM_PASSWORD_NONE) {
             bm_cairo_draw_line(cairo, &paint, &result, "%s", filter_text);
@@ -396,18 +410,18 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct
         }
 
         const uint32_t page = (menu->index / lines) * lines;
-        
+
         for (uint32_t l = 0, i = page; l < lines && posy < max_height; ++i, ++l) {
             if (!menu->fixed_height && i >= count) {
                 continue;
             }
 
-
             uint32_t is_fixed_up = (menu->fixed_height && menu->lines_mode == BM_LINES_UP);
             uint32_t required_empty = (count > lines ? 0 : lines - count);
-            int32_t last_item_index = (count < lines ? count - 1 + page : lines - 1 + page);
-            int32_t display_item_index = (menu->lines_mode == BM_LINES_DOWN ? i : menu->fixed_height ? last_item_index - i + page + required_empty  : last_item_index - i + page);
-            display_item_index = (display_item_index < 0 ? 0 : display_item_index);
+            uint32_t last_item_index = (count < lines ? count - 1 + page : lines - 1 + page);
+            last_item_index = MAX(MIN(last_item_index, count - 1), 0);
+            uint32_t display_item_index = (menu->lines_mode == BM_LINES_DOWN ? i : menu->fixed_height ? last_item_index - i + page + required_empty  : last_item_index - i + page);
+            display_item_index = MAX(MIN(display_item_index, count - 1), 0);
 
             bool highlighted = false;
             if ((i < count && !is_fixed_up) || (is_fixed_up && display_item_index <= last_item_index)) {
@@ -432,10 +446,10 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct
                 bm_cairo_color_from_menu_color(menu, BM_COLOR_ITEM_FG, &paint.fg);
                 bm_cairo_color_from_menu_color(menu, BM_COLOR_ITEM_BG, &paint.bg);
             }
-            
+
             char *line_str = "";
             if ((i < count && !is_fixed_up) || (is_fixed_up && display_item_index <= last_item_index)) {
-                line_str = bm_cairo_entry_message(items[display_item_index]->text, highlighted, menu->event_feedback, i,  count);
+                line_str = bm_cairo_entry_message(items[display_item_index]->text, highlighted, menu->event_feedback, i, count);
             }
 
             if (menu->prefix && highlighted) {
@@ -491,7 +505,7 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct
             cl += result.x_advance + 1;
         }
 
-        for (uint32_t i = menu->index; i < count && cl < (width/cairo->scale); ++i) { 
+        for (uint32_t i = menu->index; i < count && cl < (width/cairo->scale); ++i) {
             bool highlighted = (items[i] == bm_menu_get_highlighted_item(menu));
 
             if (highlighted) {
@@ -524,14 +538,14 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct
             paint.box = (struct box){ 1, 2, vpadding, -vpadding, 0, height };
             bm_cairo_draw_line(cairo, &paint, &result, ">");
         }
-        
+
     }
 
     if (menu->lines_mode == BM_LINES_UP) {
         if (menu->title) {
             bm_cairo_color_from_menu_color(menu, BM_COLOR_TITLE_FG, &paint.fg);
             bm_cairo_color_from_menu_color(menu, BM_COLOR_TITLE_BG, &paint.bg);
-        
+
             paint.pos = (struct pos){ border_size + 4, posy + vpadding + border_size };
             paint.box = (struct box){ 4, 16, vpadding, -vpadding, 0, height };
             bm_cairo_draw_line(cairo, &paint, &result, "%s", menu->title);
@@ -554,9 +568,7 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct
             bm_cairo_draw_line_str(cairo, &paint, &result, "");
         } else if (menu->password == BM_PASSWORD_INDICATOR) {
             char asterisk_print[1024] = "";
-        
             for (int i = 0; i < (int)(strlen(filter_text)); ++i) asterisk_print[i] = '*';
-        
             bm_cairo_draw_line(cairo, &paint, &result, "%s", asterisk_print);
         } else if (menu->password == BM_PASSWORD_NONE) {
             bm_cairo_draw_line(cairo, &paint, &result, "%s", filter_text);
@@ -573,7 +585,7 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct
         char counter[128];
         snprintf(counter, sizeof(counter), "[%u/%u]", filtered_item_count, total_item_count);
         bm_pango_get_text_extents(cairo, &paint, &result, "%s", counter);
-        
+
         bm_cairo_color_from_menu_color(menu, BM_COLOR_ITEM_FG, &paint.fg);
         bm_cairo_color_from_menu_color(menu, BM_COLOR_ITEM_BG, &paint.bg);
         paint.pos = (struct pos){ width/cairo->scale - result.x_advance - 10, vpadding + border_size };
@@ -585,16 +597,17 @@ bm_cairo_paint(struct cairo *cairo, uint32_t width, uint32_t max_height, struct
     bm_cairo_color_from_menu_color(menu, BM_COLOR_BORDER, &paint.fg);
     cairo_set_source_rgba(cairo->cr, paint.fg.r, paint.fg.b, paint.fg.g, paint.fg.a);
     if (!border_radius) {
-        cairo_rectangle(cairo->cr, 0, 0, (width + border_size)/cairo->scale, (height * (page_length + 1)) + (2 * border_size));
+        cairo_rectangle(cairo->cr, 0, 0, (width + border_size)/cairo->scale, total_height);
     } else {
-        bm_cairo_draw_rounded_path(cairo->cr, 0, 0, (width + border_size)/cairo->scale, (height * (page_length + 1)) + (2 * border_size), border_radius);
+        bm_cairo_rounded_path(cairo->cr, 1, 1, ((width + border_size)/cairo->scale)-2, total_height-2, border_radius);
     }
     cairo_set_line_width(cairo->cr, 2 * menu->border_size);
-    
+
     cairo_stroke(cairo->cr);
 
     out_result->height += 2 * border_size;
     out_result->height *= cairo->scale;
+    cairo_reset_clip(cairo->cr);
 }
 
 #endif /* _BM_CAIRO_H */
