@@ -150,7 +150,10 @@ poll_key(const struct bm_menu *menu, unsigned int *unicode)
             return (mods & MOD_CTRL ? BM_KEY_WORD_DELETE : BM_KEY_UNICODE);
 
         case XK_y:
-            return (mods & MOD_CTRL ? BM_KEY_PASTE : BM_KEY_UNICODE);
+            return (mods & MOD_CTRL ? BM_KEY_PASTE_PRIMARY : BM_KEY_UNICODE);
+
+        case XK_Y:
+            return (mods & MOD_CTRL ? BM_KEY_PASTE_CLIPBOARD : BM_KEY_UNICODE);
 
         case XK_j:
             return (mods & MOD_ALT ? BM_KEY_DOWN : BM_KEY_UNICODE);
